@@ -317,7 +317,6 @@ dkim_canon_header_string(struct dkim_dstring *dstr, dkim_canon_t canon,
                          unsigned char *hdr, size_t hdrlen, _Bool crlf)
 {
 	_Bool space;
-	int n;
 	u_char *p;
 	u_char *tmp;
 	u_char *end;
@@ -329,8 +328,6 @@ dkim_canon_header_string(struct dkim_dstring *dstr, dkim_canon_t canon,
 	tmp = tmpbuf;
 	end = tmpbuf + sizeof tmpbuf - 1;
 
-	n = 0;
-
 	switch (canon)
 	{
 	  case DKIM_CANON_SIMPLE:
@@ -1420,7 +1417,7 @@ dkim_canon_runheaders(DKIM *dkim)
 			SHA1_Final(sha1->sha1_out, &sha1->sha1_ctx);
 
 			if (sha1->sha1_tmpbio != NULL)
-				BIO_flush(sha1->sha1_tmpbio);
+				(void) BIO_flush(sha1->sha1_tmpbio);
 
 			break;
 		  }
@@ -1434,7 +1431,7 @@ dkim_canon_runheaders(DKIM *dkim)
 			SHA256_Final(sha256->sha256_out, &sha256->sha256_ctx);
 
 			if (sha256->sha256_tmpbio != NULL)
-				BIO_flush(sha256->sha256_tmpbio);
+				(void) BIO_flush(sha256->sha256_tmpbio);
 
 			break;
 		  }
@@ -1558,7 +1555,7 @@ dkim_canon_signature(DKIM *dkim, struct dkim_header *hdr)
 			SHA1_Final(sha1->sha1_out, &sha1->sha1_ctx);
 
 			if (sha1->sha1_tmpbio != NULL)
-				BIO_flush(sha1->sha1_tmpbio);
+				(void) BIO_flush(sha1->sha1_tmpbio);
 
 			break;
 		  }
@@ -1572,7 +1569,7 @@ dkim_canon_signature(DKIM *dkim, struct dkim_header *hdr)
 			SHA256_Final(sha256->sha256_out, &sha256->sha256_ctx);
 
 			if (sha256->sha256_tmpbio != NULL)
-				BIO_flush(sha256->sha256_tmpbio);
+				(void) BIO_flush(sha256->sha256_tmpbio);
 
 			break;
 		  }
@@ -1976,7 +1973,7 @@ dkim_canon_closebody(DKIM *dkim)
 			SHA1_Final(sha1->sha1_out, &sha1->sha1_ctx);
 
 			if (sha1->sha1_tmpbio != NULL)
-				BIO_flush(sha1->sha1_tmpbio);
+				(void) BIO_flush(sha1->sha1_tmpbio);
 
 			break;
 		  }
@@ -1990,7 +1987,7 @@ dkim_canon_closebody(DKIM *dkim)
 			SHA256_Final(sha256->sha256_out, &sha256->sha256_ctx);
 
 			if (sha256->sha256_tmpbio != NULL)
-				BIO_flush(sha256->sha256_tmpbio);
+				(void) BIO_flush(sha256->sha256_tmpbio);
 
 			break;
 		  }
@@ -2076,3 +2073,49 @@ dkim_canon_getfinal(DKIM_CANON *canon, u_char **digest, size_t *dlen)
 		return DKIM_STAT_INTERNAL;
 	}
 }
+
+/*
+**  DKIM_CANON_GETHASHES -- retrieve hashes
+**
+**  Parameters:
+**  	sig -- signature from which to get completed hashes
+**  	hh -- pointer to header hash buffer (returned)
+**  	hhlen -- bytes used at hh (returned)
+**  	bh -- pointer to body hash buffer (returned)
+**  	bhlen -- bytes used at bh (returned)
+**
+**  Return value:
+**  	DKIM_STAT_OK -- successful completion
+**  	DKIM_STAT_INVALID -- hashing hasn't been completed
+*/
+
+DKIM_STAT
+dkim_canon_gethashes(DKIM_SIGINFO *sig, void **hh, size_t *hhlen,
+                     void **bh, size_t *bhlen)
+{
+	DKIM_STAT status;
+	struct dkim_canon *hdc;
+	struct dkim_canon *bdc;
+	u_char *hd;
+	u_char *bd;
+	size_t hdlen;
+	size_t bdlen;
+
+	hdc = sig->sig_hdrcanon;
+	bdc = sig->sig_bodycanon;
+
+	status = dkim_canon_getfinal(hdc, &hd, &hdlen);
+	if (status != DKIM_STAT_OK)
+		return status;
+
+	status = dkim_canon_getfinal(bdc, &bd, &bdlen);
+	if (status != DKIM_STAT_OK)
+		return status;
+
+	*hh = hd;
+	*hhlen = hdlen;
+	*bh = bd;
+	*bhlen = bdlen;
+
+	return DKIM_STAT_OK;
+}
