@@ -1,11 +1,11 @@
 /*
-**  Copyright (c) 2009, 2010, The OpenDKIM Project.  All rights reserved.
+**  Copyright (c) 2009-2011, The OpenDKIM Project.  All rights reserved.
 **
-**  $Id: miltertest.c,v 1.18.2.1 2010/07/11 09:35:22 cm-msk Exp $
+**  $Id: miltertest.c,v 1.42 2010/09/23 18:37:32 cm-msk Exp $
 */
 
 #ifndef lint
-static char miltertest_c_id[] = "$Id: miltertest.c,v 1.18.2.1 2010/07/11 09:35:22 cm-msk Exp $";
+static char miltertest_c_id[] = "$Id: miltertest.c,v 1.42 2010/09/23 18:37:32 cm-msk Exp $";
 #endif /* ! lint */
 
 #include "build-config.h"
@@ -18,6 +18,8 @@ static char miltertest_c_id[] = "$Id: miltertest.c,v 1.18.2.1 2010/07/11 09:35:2
 #include <sys/wait.h>
 #include <sys/un.h>
 #include <sys/uio.h>
+#include <sys/time.h>
+#include <sys/resource.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <sysexits.h>
@@ -47,6 +49,11 @@ static char miltertest_c_id[] = "$Id: miltertest.c,v 1.18.2.1 2010/07/11 09:35:2
 #include <lualib.h>
 #include <lauxlib.h>
 
+/* types */
+#ifndef HAVE_USECONDS_T
+typedef unsigned int useconds_t;
+#endif /* ! HAVE_USECONDS_T */
+
 /* macros */
 #ifndef FALSE
 # define FALSE			0
@@ -58,7 +65,7 @@ static char miltertest_c_id[] = "$Id: miltertest.c,v 1.18.2.1 2010/07/11 09:35:2
 #ifdef SMFIP_NODATA
 # define CHECK_MPOPTS(c,o)	(((c)->ctx_mpopts & (o)) != 0)
 #else /* SMFIP_NODATA */
-# define CHECK_MPOPTS()		0
+# define CHECK_MPOPTS(c,o)	0
 #endif /* SMFIP_NODATA */
 
 #ifndef SMFIP_NR_CONN
@@ -74,12 +81,12 @@ static char miltertest_c_id[] = "$Id: miltertest.c,v 1.18.2.1 2010/07/11 09:35:2
 #endif /* SMFIP_NR_CONN */
 
 #define	MT_PRODUCT		"OpenDKIM milter test facility"
-#define	MT_VERSION		"1.0.0"
+#define	MT_VERSION		"1.4.0"
 
 #define	BUFRSZ			1024
 #define	CHUNKSZ			65536
 
-#define	CMDLINEOPTS		"D:s:vV"
+#define	CMDLINEOPTS		"D:s:uvVw"
 
 #define	DEFBODY			"Dummy message body.\r\n"
 #define	DEFCLIENTPORT		12345
@@ -138,6 +145,7 @@ int mt_mailfrom(lua_State *);
 int mt_negotiate(lua_State *);
 int mt_rcptto(lua_State *);
 int mt_set_timeout(lua_State *);
+int mt_signal(lua_State *);
 int mt_sleep(lua_State *);
 int mt_startfilter(lua_State *);
 int mt_test_action(lua_State *);
@@ -195,6 +203,7 @@ static const luaL_Reg mt_library[] =
 	{ "negotiate",		mt_negotiate	},
 	{ "rcptto",		mt_rcptto	},
 	{ "set_timeout",	mt_set_timeout	},
+	{ "signal",		mt_signal	},
 	{ "sleep",		mt_sleep	},
 	{ "startfilter",	mt_startfilter	},
 	{ "test_action",	mt_test_action	},
@@ -203,8 +212,9 @@ static const luaL_Reg mt_library[] =
 	{ NULL,			NULL 		}
 };
 
-
 /* globals */
+_Bool rusage;
+_Bool nowait;
 int verbose;
 unsigned int tmo;
 pid_t filterpid;
@@ -613,6 +623,16 @@ mt_assert_state(struct mt_context *ctx, int state)
 			return FALSE;
 		}
 
+		/* decode and store requested protocol steps and actions */
+		(void) memcpy((char *) &nvers, buf, MILTER_LEN_BYTES);
+		(void) memcpy((char *) &nacts, buf + MILTER_LEN_BYTES,
+		              MILTER_LEN_BYTES);
+		(void) memcpy((char *) &npopts, buf + (MILTER_LEN_BYTES * 2),
+		              MILTER_LEN_BYTES);
+
+		ctx->ctx_mactions = ntohl(nacts);
+		ctx->ctx_mpopts = ntohl(npopts);
+
 		ctx->ctx_state = STATE_NEGOTIATED;
 	}
 
@@ -1058,7 +1078,11 @@ mt_getcwd(lua_State *l)
 
 	memset(dir, '\0', sizeof dir);
 
-	(void) getcwd(dir, MAXPATHLEN);
+	if (getcwd(dir, MAXPATHLEN) == NULL)
+	{
+		lua_pushstring(l, "mt.getcwd(): getcwd returned error");
+		lua_error(l);
+	}
 
 	lua_pushstring(l, dir);
 
@@ -1204,7 +1228,7 @@ mt_startfilter(lua_State *l)
 
 		if (verbose > 0)
 		{
-			fprintf(stderr, "%s: `%s' started in process %d\n",
+			fprintf(stderr, "%s: '%s' started in process %d\n",
 			        progname, argv[0], filterpid);
 		}
 
@@ -1218,6 +1242,52 @@ mt_startfilter(lua_State *l)
 	return 1;
 }
 
+/*
+**  MT_SIGNAL -- signal a filter
+**
+**  Parameters:
+**  	l -- Lua state
+**
+**  Return value:
+**   	nil (on the Lua stack)
+*/
+
+int
+mt_signal(lua_State *l)
+{
+	int signum;
+
+	assert(l != NULL);
+
+	if (lua_gettop(l) != 1 || !lua_isnumber(l, 1))
+	{
+		lua_pushstring(l, "mt.signal(): Invalid argument");
+		lua_error(l);
+	}
+
+	signum = lua_tonumber(l, 1);
+	lua_pop(l, 1);
+
+	if (filterpid <= 1)
+	{
+		lua_pushstring(l, "mt.signal(): Filter not running");
+		lua_error(l);
+	}
+
+	if (kill(filterpid, signum) != 0)
+	{
+		lua_pushfstring(l, "mt.signal(): kill(): %s", strerror(errno));
+		lua_error(l);
+	}
+
+	if (verbose > 0)
+		fprintf(stderr, "%s: sent signal %d\n", progname, signum);
+
+	lua_pushnil(l);
+
+	return 1;
+}
+
 /*
 **  MT_CONNECT -- connect to a filter, returning a handle
 **
@@ -1231,8 +1301,12 @@ mt_startfilter(lua_State *l)
 int
 mt_connect(lua_State *l)
 {
+	int top;
 	int af;
-	int fd;
+	int fd = -1;
+	int saverr = 0;
+	u_int count = 1;
+	useconds_t interval = 0;
 	char *at;
 	char *p;
 	const char *sockinfo;
@@ -1240,15 +1314,23 @@ mt_connect(lua_State *l)
 
 	assert(l != NULL);
 
-	if (lua_gettop(l) != 1 ||
-	    !lua_isstring(l, 1))
+	top = lua_gettop(l);
+
+	if (!(top == 1 && lua_isstring(l, 1)) &&
+	    !(top == 3 && lua_isstring(l, 1) && lua_isnumber(l, 2) &&
+	                  lua_isnumber(l, 3)))
 	{
 		lua_pushstring(l, "mt.connect(): Invalid argument");
 		lua_error(l);
 	}
 
 	sockinfo = lua_tostring(l, 1);
-	lua_pop(l, 1);
+	if (top == 3)
+	{
+		count = (u_int) lua_tonumber(l, 2);
+		interval = (useconds_t) (1000000. * lua_tonumber(l, 3));
+	}
+	lua_pop(l, top);
 
 	af = AF_UNSPEC;
 	p = strchr(sockinfo, ':');
@@ -1294,10 +1376,33 @@ mt_connect(lua_State *l)
 			lua_error(l);
 		}
 
-		if (connect(fd, (struct sockaddr *) &sa, sizeof sa) < 0)
+		while (count > 0)
 		{
-			lua_pushfstring(l, "mt.connect(): connect(): %s",
-			                strerror(errno));
+			saverr = 0;
+
+			if (connect(fd, (struct sockaddr *) &sa,
+			            sizeof sa) == 0)
+				break;
+
+			saverr = errno;
+
+			if (verbose > 1)
+			{
+				fprintf(stdout,
+				        "%s: connect(): %s; %u tr%s left\n",
+				        progname, strerror(errno), count - 1,
+				        count == 2 ? "y" : "ies");
+			}
+
+			usleep(interval);
+
+			count--;
+		}
+
+		if (saverr != 0)
+		{
+			lua_pushfstring(l, "mt.connect(): %s: connect(): %s",
+			                sockinfo, strerror(errno));
 			lua_error(l);
 		}
 
@@ -1369,10 +1474,33 @@ mt_connect(lua_State *l)
 			lua_error(l);
 		}
 
-		if (connect(fd, (struct sockaddr *) &sa, sizeof sa) < 0)
+		while (count > 0)
 		{
-			lua_pushfstring(l, "mt.connect(): connect(): %s",
-			                strerror(errno));
+			saverr = 0;
+
+			if (connect(fd, (struct sockaddr *) &sa,
+			            sizeof sa) == 0)
+				break;
+
+			saverr = errno;
+
+			if (verbose > 1)
+			{
+				fprintf(stdout,
+				        "%s: connect(): %s; %u tr%s left\n",
+				        progname, strerror(errno), count - 1,
+				        count == 2 ? "y" : "ies");
+			}
+
+			usleep(interval);
+
+			count--;
+		}
+
+		if (saverr != 0)
+		{
+			lua_pushfstring(l, "mt.connect(): %s: connect(): %s",
+			                sockinfo, strerror(errno));
 			lua_error(l);
 		}
 
@@ -1402,7 +1530,7 @@ mt_connect(lua_State *l)
 
 	if (verbose > 0)
 	{
-		fprintf(stdout, "%s: connected to `%s', fd %d\n",
+		fprintf(stdout, "%s: connected to '%s', fd %d\n",
 		        progname, sockinfo, fd);
 	}
 
@@ -1422,7 +1550,8 @@ mt_connect(lua_State *l)
 int
 mt_sleep(lua_State *l)
 {
-	int secs;
+	double p;
+	useconds_t usecs;
 
 	assert(l != NULL);
 
@@ -1433,16 +1562,17 @@ mt_sleep(lua_State *l)
 		lua_error(l);
 	}
 
-	secs = lua_tonumber(l, 1);
+	p = lua_tonumber(l, 1);
+	usecs = (useconds_t) (1000000. * p);
 	lua_pop(l, 1);
 
 	if (verbose > 1)
 	{
-		fprintf(stdout, "%s: pausing for %d second%s\n",
-		        progname, secs, secs == 1 ? "" : "s");
+		fprintf(stdout, "%s: pausing for %f second%s\n",
+		        progname, p, usecs == 1000000 ? "" : "s");
 	}
 
-	sleep(secs);
+	usleep(usecs);
 
 	lua_pushnil(l);
 
@@ -1679,7 +1809,7 @@ mt_negotiate(lua_State *l)
 	if (verbose > 0)
 	{
 		fprintf(stdout,
-		        "%s: option negotiation sent on fd %d, reply `%c'\n",
+		        "%s: option negotiation sent on fd %d, reply '%c'\n",
 		        progname, ctx->ctx_fd, rcmd);
 	}
 
@@ -1774,7 +1904,7 @@ mt_macro(lua_State *l)
 
 	if (verbose > 0)
 	{
-		fprintf(stdout, "%s: %d `%c' macro(s) sent on fd %d\n",
+		fprintf(stdout, "%s: %d '%c' macro(s) sent on fd %d\n",
 		        progname, n, type, ctx->ctx_fd);
 	}
 
@@ -1797,7 +1927,7 @@ int
 mt_conninfo(lua_State *l)
 {
 	char rcmd;
-	char family;
+	char family = 'U';
 	size_t buflen;
 	size_t s;
 	uint16_t port;
@@ -1805,7 +1935,6 @@ mt_conninfo(lua_State *l)
 	char *host;
 	char *bp;
 	char *ipstr;
-	struct in_addr sa;
 	char buf[BUFRSZ];
 	char tmp[BUFRSZ];
 
@@ -1844,14 +1973,14 @@ mt_conninfo(lua_State *l)
 	if (ipstr == NULL)
 	{
 #if (HAVE_GETADDRINFO && HAVE_INET_NTOP)
-		char *a;
+		char *a = NULL;
 		struct addrinfo *res;
 		struct sockaddr_in *s4;
 		struct sockaddr_in6 *s6;
 
 		if (getaddrinfo(host, NULL, NULL, &res) != 0)
 		{
-			lua_pushfstring(l, "mt.conninfo(): host `%s' unknown",
+			lua_pushfstring(l, "mt.conninfo(): host '%s' unknown",
 			                host);
 			lua_error(l);
 		}
@@ -1869,26 +1998,31 @@ mt_conninfo(lua_State *l)
 			family = '6';
 		}
 
-		memset(tmp, '\0', sizeof tmp);
-
-		if (inet_ntop(res->ai_family, a, tmp, sizeof tmp - 1) == NULL)
+		if (family != 'U')
 		{
-			freeaddrinfo(res);
-			lua_pushfstring(l,
-			                "mt.conninfo(): can't convert address for host `%s' to text",
-			                host);
-			lua_error(l);
+			memset(tmp, '\0', sizeof tmp);
+
+			if (inet_ntop(res->ai_family, a,
+			              tmp, sizeof tmp - 1) == NULL)
+			{
+				freeaddrinfo(res);
+				lua_pushfstring(l,
+				                "mt.conninfo(): can't convert address for host '%s' to text",
+				                host);
+				lua_error(l);
+			}
 		}
 
 		freeaddrinfo(res);
 		ipstr = tmp;
 #else /* HAVE_GETADDRINFO && HAVE_INET_NTOP */
 		struct hostent *h;
+		struct in_addr sa;
 
 		h = gethostbyname(host);
 		if (h == NULL)
 		{
-			lua_pushfstring(l, "mt.conninfo(): host `%s' unknown",
+			lua_pushfstring(l, "mt.conninfo(): host '%s' unknown",
 			                host);
 			lua_error(l);
 		}
@@ -1899,7 +2033,7 @@ mt_conninfo(lua_State *l)
 		family = '4';
 #endif /* HAVE_GETADDRINFO && HAVE_INET_NTOP */
 	}
-	else
+	else if (strcasecmp(ipstr, "unspec") != 0)
 	{
 #ifdef HAVE_INET_PTON
 		struct in_addr a;
@@ -1916,16 +2050,18 @@ mt_conninfo(lua_State *l)
 		else
 		{
 			lua_pushfstring(l,
-			                "mt.conninfo(): invalid IP address `%s'",
+			                "mt.conninfo(): invalid IP address '%s'",
 			                ipstr);
 			lua_error(l);
 		}
 #else /* HAVE_INET_PTON */
+		struct in_addr sa;
+
 		sa.s_addr = inet_addr(ipstr);
 		if (sa.s_addr == INADDR_NONE)
 		{
 			lua_pushfstring(l,
-			                "mt.conninfo(): invalid IPv4 address `%s'",
+			                "mt.conninfo(): invalid IPv4 address '%s'",
 			                ipstr);
 			lua_error(l);
 		}
@@ -1933,19 +2069,25 @@ mt_conninfo(lua_State *l)
 #endif /* HAVE_INET_PTON */
 	}
 
-	s = strlen(host) + 1 + sizeof(char) + sizeof port + strlen(ipstr) + 1;
-
-	port = htons(DEFCLIENTPORT);		/* don't really need this */
-
 	bp = buf;
 	memcpy(bp, host, strlen(host));
 	bp += strlen(host);
 	*bp++ = '\0';
 	memcpy(bp, &family, sizeof family);
 	bp += sizeof family;
-	memcpy(bp, &port, sizeof port);
-	bp += sizeof port;
-	memcpy(bp, ipstr, strlen(ipstr) + 1);
+
+	s = strlen(host) + 1 + sizeof(char);
+
+	if (family != 'U')			/* known family data */
+	{
+		port = htons(DEFCLIENTPORT);	/* don't really need this */
+
+		memcpy(bp, &port, sizeof port);
+		bp += sizeof port;
+		memcpy(bp, ipstr, strlen(ipstr) + 1);
+
+		s += sizeof port + strlen(ipstr) + 1;
+	}
 
 	if (!mt_milter_write(ctx->ctx_fd, SMFIC_CONNECT, buf, s))
 	{
@@ -1970,7 +2112,7 @@ mt_conninfo(lua_State *l)
 	if (verbose > 0)
 	{
 		fprintf(stdout,
-		        "%s: connection details sent on fd %d, reply `%c'\n",
+		        "%s: connection details sent on fd %d, reply '%c'\n",
 		        progname, ctx->ctx_fd, rcmd);
 	}
 
@@ -1992,6 +2134,7 @@ mt_conninfo(lua_State *l)
 int
 mt_unknown(lua_State *l)
 {
+#ifdef SMFIC_UNKNOWN
 	char rcmd;
 	size_t buflen;
 	size_t s;
@@ -1999,6 +2142,7 @@ mt_unknown(lua_State *l)
 	char *cmd;
 	char *bp;
 	char buf[BUFRSZ];
+#endif /* SMFIC_UNKNOWN */
 
 	assert(l != NULL);
 
@@ -2058,14 +2202,14 @@ mt_unknown(lua_State *l)
 	if (verbose > 0)
 	{
 		fprintf(stdout,
-		        "%s: UNKNOWN sent on fd %d, reply `%c'\n",
+		        "%s: UNKNOWN sent on fd %d, reply '%c'\n",
 		        progname, ctx->ctx_fd, rcmd);
 	}
 
 	lua_pushnil(l);
+#endif /* ! SMFIC_UNKNOWN */
 
 	return 1;
-#endif /* ! SMFIC_UNKNOWN */
 }
 
 /*
@@ -2144,7 +2288,7 @@ mt_helo(lua_State *l)
 	if (verbose > 0)
 	{
 		fprintf(stdout,
-		        "%s: HELO sent on fd %d, reply `%c'\n",
+		        "%s: HELO sent on fd %d, reply '%c'\n",
 		        progname, ctx->ctx_fd, rcmd);
 	}
 
@@ -2239,7 +2383,7 @@ mt_mailfrom(lua_State *l)
 	if (verbose > 0)
 	{
 		fprintf(stdout,
-		        "%s: MAIL sent on fd %d, reply `%c'\n",
+		        "%s: MAIL sent on fd %d, reply '%c'\n",
 		        progname, ctx->ctx_fd, rcmd);
 	}
 
@@ -2340,7 +2484,7 @@ mt_rcptto(lua_State *l)
 	if (verbose > 0)
 	{
 		fprintf(stdout,
-		        "%s: RCPT sent on fd %d, reply `%c'\n",
+		        "%s: RCPT sent on fd %d, reply '%c'\n",
 		        progname, ctx->ctx_fd, rcmd);
 	}
 
@@ -2362,10 +2506,12 @@ mt_rcptto(lua_State *l)
 int
 mt_data(lua_State *l)
 {
+#ifdef SMFIC_DATA
 	char rcmd;
 	size_t buflen;
 	struct mt_context *ctx;
 	char buf[BUFRSZ];
+#endif /* SMFIC_DATA */
 
 	assert(l != NULL);
 
@@ -2417,14 +2563,14 @@ mt_data(lua_State *l)
 	if (verbose > 0)
 	{
 		fprintf(stdout,
-		        "%s: DATA sent on fd %d, reply `%c'\n",
+		        "%s: DATA sent on fd %d, reply '%c'\n",
 		        progname, ctx->ctx_fd, rcmd);
 	}
 
 	lua_pushnil(l);
+#endif /* ! SMFIC_DATA */
 
 	return 1;
-#endif /* ! SMFIC_DATA */
 }
 
 /*
@@ -2466,10 +2612,18 @@ mt_header(lua_State *l)
 	lua_pop(l, 3);
 
 	s = strlen(name) + 1 + strlen(value) + 1;
+#ifdef SMFIP_HDR_LEADSPC
+	if (CHECK_MPOPTS(ctx, SMFIP_HDR_LEADSPC))
+		s++;
+#endif /* SMFIP_HDR_LEADSPC */
 
 	bp = buf;
 	memcpy(buf, name, strlen(name) + 1);
 	bp += strlen(name) + 1;
+#ifdef SMFIP_HDR_LEADSPC
+	if (CHECK_MPOPTS(ctx, SMFIP_HDR_LEADSPC))
+		*bp++ = ' ';
+#endif /* SMFIP_HDR_LEADSPC */
 	memcpy(bp, value, strlen(value) + 1);
 
 	if (!mt_assert_state(ctx, STATE_ENVRCPT))
@@ -2506,7 +2660,7 @@ mt_header(lua_State *l)
 	if (verbose > 0)
 	{
 		fprintf(stdout,
-		        "%s: header sent on fd %d, reply `%c'\n",
+		        "%s: header sent on fd %d, reply '%c'\n",
 		        progname, ctx->ctx_fd, rcmd);
 	}
 
@@ -2579,7 +2733,7 @@ mt_eoh(lua_State *l)
 	if (verbose > 0)
 	{
 		fprintf(stdout,
-		        "%s: EOH sent on fd %d, reply `%c'\n",
+		        "%s: EOH sent on fd %d, reply '%c'\n",
 		        progname, ctx->ctx_fd, rcmd);
 	}
 
@@ -2655,7 +2809,7 @@ mt_bodystring(lua_State *l)
 	if (verbose > 0)
 	{
 		fprintf(stdout,
-		        "%s: %zu byte(s) of body sent on fd %d, reply `%c'\n",
+		        "%s: %lu byte(s) of body sent on fd %d, reply '%c'\n",
 		        progname, strlen(str), ctx->ctx_fd, rcmd);
 	}
 
@@ -2746,7 +2900,7 @@ mt_bodyrandom(lua_State *l)
 		if (verbose > 0)
 		{
 			fprintf(stdout,
-			        "%s: %zu byte(s) of body sent on fd %d, reply `%c'\n",
+			        "%s: %lu byte(s) of body sent on fd %d, reply '%c'\n",
 			        progname, strlen(buf), ctx->ctx_fd, rcmd);
 		}
 
@@ -2847,7 +3001,7 @@ mt_bodyfile(lua_State *l)
 			if (verbose > 0)
 			{
 				fprintf(stdout,
-				        "%s: %zu byte(s) of body sent on fd %d, reply `%c'\n",
+				        "%s: %lu byte(s) of body sent on fd %d, reply '%c'\n",
 				        progname, rlen, ctx->ctx_fd, rcmd);
 			}
 		}
@@ -2940,7 +3094,7 @@ mt_eom(lua_State *l)
 
 	if (verbose > 0)
 	{
-		fprintf(stdout, "%s: EOM sent on fd %d, reply `%c'\n",
+		fprintf(stdout, "%s: EOM sent on fd %d, reply '%c'\n",
 		        progname, ctx->ctx_fd, rcmd);
 	}
 
@@ -3045,6 +3199,7 @@ mt_eom_check(lua_State *l)
 
 	  case MT_HDRINSERT:
 	  {
+#ifdef SMFIR_INSHEADER
 		int idx = -1;
 		char *name = NULL;
 		char *value = NULL;
@@ -3112,6 +3267,7 @@ mt_eom_check(lua_State *l)
 				}
 			}
 		}
+#endif /* SMFIR_INSHEADER */
 
 		lua_pushboolean(l, 0);
 		return 1;
@@ -3365,6 +3521,7 @@ mt_eom_check(lua_State *l)
 		return 1;
 	  }
 
+#ifdef SMFIR_QUARANTINE
 	  case MT_QUARANTINE:
 	  {
 		char *reason = NULL;
@@ -3401,6 +3558,7 @@ mt_eom_check(lua_State *l)
 		lua_pushboolean(l, 0);
 		return 1;
 	  }
+#endif /* SMFIR_QUARANTINE */
 
 	  case MT_SMTPREPLY:
 	  {
@@ -3587,12 +3745,17 @@ mt_getheader(lua_State *l)
 
 	for (r = ctx->ctx_eomreqs; r != NULL; r = r->eom_next)
 	{
+#ifdef SMFIR_INSHEADER
 		if (r->eom_request == SMFIR_ADDHEADER ||
 		    r->eom_request == SMFIR_INSHEADER)
+#else /* SMFIR_INSHEADER */
+		if (r->eom_request == SMFIR_ADDHEADER)
+#endif /* SMFIR_INSHEADER */
 		{
 			char *rname;
 			char *rvalue;
 
+#ifdef SMFIR_INSHEADER
 			if (r->eom_request == SMFIR_INSHEADER)
 			{
 				rname = r->eom_rdata + MILTER_LEN_BYTES;
@@ -3600,6 +3763,7 @@ mt_getheader(lua_State *l)
 				         strlen(rname) + 1;
 			}
 			else
+#endif /* SMFIR_INSHEADER */
 			{
 				rname = r->eom_rdata;
 				rvalue = r->eom_rdata + strlen(rname) + 1;
@@ -3641,7 +3805,10 @@ usage(void)
 	fprintf(stderr, "%s: usage: %s [options]\n"
 	                "\t-D name[=value]\tdefine global variable\n"
 	                "\t-s script      \tscript to run (default = stdin)\n"
-	                "\t-v             \tverbose mode\n",
+	                "\t-u             \treport usage statistics\n"
+	                "\t-v             \tverbose mode\n"
+	                "\t-V             \tprint version number and exit\n"
+	                "\t-w             \tdon't wait for child at shutdown\n",
 	                progname, progname);
 
 	return EX_USAGE;
@@ -3676,6 +3843,8 @@ main(int argc, char **argv)
 	verbose = 0;
 	filterpid = 0;
 	tmo = DEFTIMEOUT;
+	rusage = FALSE;
+	nowait = FALSE;
 
 	l = lua_newstate(mt_lua_alloc, NULL);
 	if (l == NULL)
@@ -3711,7 +3880,7 @@ main(int argc, char **argv)
 			if (script != NULL)
 			{
 				fprintf(stderr,
-				        "%s: multiple use of `-%c' not permitted\n",
+				        "%s: multiple use of '-%c' not permitted\n",
 				        progname, c);
 				lua_close(l);
 				return EX_USAGE;
@@ -3720,6 +3889,10 @@ main(int argc, char **argv)
 			script = optarg;
 			break;
 
+		  case 'u':
+			rusage = TRUE;
+			break;
+
 		  case 'v':
 			verbose++;
 			break;
@@ -3729,6 +3902,10 @@ main(int argc, char **argv)
 			        MT_VERSION);
 			return 0;
 
+		  case 'w':
+			nowait = TRUE;
+			break;
+
 		  default:
 			lua_close(l);
 			return usage();
@@ -3777,7 +3954,7 @@ main(int argc, char **argv)
 		if (rlen != s.st_size)
 		{
 			fprintf(stderr,
-			        "%s: %s: read() returned %zu (expecting %ld)\n",
+			        "%s: %s: read() returned %lu (expecting %ld)\n",
 			        progname, script, rlen, s.st_size);
 			free((void *) io.lua_io_script);
 			close(fd);
@@ -3866,10 +4043,14 @@ main(int argc, char **argv)
 	lua_pushnumber(l, SMFIP_NOHREPL);
 	lua_setglobal(l, "SMFIP_NOHREPL");
 #endif /* SMFIP_NOHREPL */
+#ifdef SMFIP_NOUNKNOWN
 	lua_pushnumber(l, SMFIP_NOUNKNOWN);
 	lua_setglobal(l, "SMFIP_NOUNKNOWN");
+#endif /* SMFIP_NOUNKNOWN */
+#ifdef SMFIP_NODATA
 	lua_pushnumber(l, SMFIP_NODATA);
 	lua_setglobal(l, "SMFIP_NODATA");
+#endif /* SMFIP_NODATA */
 #ifdef SMFIP_SKIP
 	lua_pushnumber(l, SMFIP_SKIP);
 	lua_setglobal(l, "SMFIP_SKIP");
@@ -3923,8 +4104,10 @@ main(int argc, char **argv)
 	lua_setglobal(l, "SMFIF_DELRCPT");
 	lua_pushnumber(l, SMFIF_CHGHDRS);
 	lua_setglobal(l, "SMFIF_CHGHDRS");
+#ifdef SMFIF_QUARANTINE
 	lua_pushnumber(l, SMFIF_QUARANTINE);
 	lua_setglobal(l, "SMFIF_QUARANTINE");
+#endif /* SMFIF_QUARANTINE */
 #ifdef SMFIF_CHGFROM
 	lua_pushnumber(l, SMFIF_CHGFROM);
 	lua_setglobal(l, "SMFIF_CHGFROM");
@@ -3971,6 +4154,27 @@ main(int argc, char **argv)
 		        lua_tostring(l, 1));
 	}
 
+	if (rusage)
+	{
+		struct rusage u;
+
+		if (getrusage(RUSAGE_SELF, &u) != 0)
+		{
+			fprintf(stderr,
+			        "%s: getrusage(RUSAGE_SELF): %s\n",
+			        progname, strerror(errno));
+
+			retval = 2;
+		}
+
+		fprintf(stdout, "%s: self:  user %u.%06u, system %u.%06u\n",
+		        progname,
+		        (unsigned) u.ru_utime.tv_sec,
+		        (unsigned) u.ru_utime.tv_usec,
+		        (unsigned) u.ru_stime.tv_sec,
+		        (unsigned) u.ru_stime.tv_usec);
+	}
+
 	if (status != 0)
 		retval = 1;
 
@@ -3985,7 +4189,7 @@ main(int argc, char **argv)
 			fprintf(stderr, "%s: %d: kill() %s\n", progname,
 			        filterpid, strerror(errno));
 		}
-		else
+		else if (!nowait)
 		{
 			if (verbose > 1)
 			{
@@ -4016,5 +4220,26 @@ main(int argc, char **argv)
 		}
 	}
 
+	if (rusage && !nowait)
+	{
+		struct rusage u;
+
+		if (getrusage(RUSAGE_CHILDREN, &u) != 0)
+		{
+			fprintf(stderr,
+			        "%s: getrusage(RUSAGE_CHILDREN): %s\n",
+			        progname, strerror(errno));
+
+			retval = 2;
+		}
+
+		fprintf(stdout, "%s: child: user %u.%06u, system %u.%06u\n",
+		        progname,
+		        (unsigned) u.ru_utime.tv_sec,
+		        (unsigned) u.ru_utime.tv_usec,
+		        (unsigned) u.ru_stime.tv_sec,
+		        (unsigned) u.ru_stime.tv_usec);
+	}
+
 	return retval;
 }
