@@ -2,18 +2,23 @@
 **  Copyright (c) 2007-2009 Sendmail, Inc. and its suppliers.
 **	All rights reserved.
 **
-**  Copyright (c) 2009, The OpenDKIM Project.  All rights reserved.
+**  Copyright (c) 2009, 2011, The OpenDKIM Project.  All rights reserved.
 **
-**  $Id: opendkim-ar.c,v 1.5 2009/10/07 18:06:17 cm-msk Exp $
+**  $Id: opendkim-ar.c,v 1.5.58.1 2010/10/27 21:43:09 cm-msk Exp $
 */
 
 #ifndef lint
-static char opendkim_ar_c_id[] = "@(#)$Id: opendkim-ar.c,v 1.5 2009/10/07 18:06:17 cm-msk Exp $";
+static char opendkim_ar_c_id[] = "@(#)$Id: opendkim-ar.c,v 1.5.58.1 2010/10/27 21:43:09 cm-msk Exp $";
 #endif /* !lint */
 
+#include "build-config.h"
+
 /* system includes */
 #include <sys/types.h>
 #include <sys/param.h>
+#ifdef HAVE_STDBOOL_H
+# include <stdbool.h>
+#endif /* HAVE_STDBOOL_H */
 #include <ctype.h>
 #include <assert.h>
 #include <string.h>
@@ -102,9 +107,9 @@ static int
 ares_tokenize(u_char *input, u_char *outbuf, size_t outbuflen,
               u_char **tokens, int ntokens)
 {
-	bool quoted = FALSE;
-	bool escaped = FALSE;
-	bool intok = FALSE;
+	_Bool quoted = FALSE;
+	_Bool escaped = FALSE;
+	_Bool intok = FALSE;
 	int n = 0;
 	int parens = 0;
 	u_char *p;
@@ -139,7 +144,7 @@ ares_tokenize(u_char *input, u_char *outbuf, size_t outbuflen,
 		{
 			escaped = TRUE;
 		}
-		else if (*p == '"')			/* quoting */
+		else if (*p == '"' && parens == 0)	/* quoting */
 		{
 			quoted = !quoted;
 
@@ -150,7 +155,7 @@ ares_tokenize(u_char *input, u_char *outbuf, size_t outbuflen,
 				intok = TRUE;
 			}
 		}
-		else if (*p == '(')			/* "(" (comment) */
+		else if (*p == '(' && !quoted)		/* "(" (comment) */
 		{
 			parens++;
 
@@ -165,21 +170,24 @@ ares_tokenize(u_char *input, u_char *outbuf, size_t outbuflen,
 			q++;
 
 		}
-		else if (*p == ')')			/* ")" (comment) */
+		else if (*p == ')' && !quoted)		/* ")" (comment) */
 		{
-			parens--;
-
-			if (parens == 0)
+			if (parens > 0)
 			{
-				intok = FALSE;
-				n++;
+				parens--;
 
-				*q = ')';
-				q++;
-				if (q <= end)
+				if (parens == 0)
 				{
-					*q = '\0';
+					intok = FALSE;
+					n++;
+
+					*q = ')';
 					q++;
+					if (q <= end)
+					{
+						*q = '\0';
+						q++;
+					}
 				}
 			}
 		}
@@ -376,7 +384,7 @@ ares_xconvert(struct lookup *table, int code)
 int
 ares_parse(u_char *hdr, struct authres *ar)
 {
-	bool quoted;
+	_Bool quoted;
 	int n;
 	int ntoks;
 	int c;
@@ -672,7 +680,7 @@ main(int argc, char **argv)
 
 	c = ares_tokenize(argv[1], buf, sizeof buf, toks, NTOKENS);
 	for (d = 0; d < c; d++)
-		printf("token %d = `%s'\n", d, toks[d]);
+		printf("token %d = '%s'\n", d, toks[d]);
 
 	printf("\n");
 
@@ -686,8 +694,8 @@ main(int argc, char **argv)
 	printf("%d result%s found\n", ar.ares_count,
 	       ar.ares_count == 1 ? "" : "s");
 
-	printf("authserv-id `%s'\n", ar.ares_host);
-	printf("version `%s'\n", ar.ares_version);
+	printf("authserv-id '%s'\n", ar.ares_host);
+	printf("version '%s'\n", ar.ares_version);
 
 	for (c = 0; c < ar.ares_count; c++)
 	{
